/*
 * vala-panel-appmenu
 * Copyright (C) 2018 Konstantin Pugin <ria.freelander@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef REGISTRARDBUSMENU_H
#define REGISTRARDBUSMENU_H

#include <gio/gio.h>

#define DBUSMENU_REG_IFACE "com.canonical.AppMenu.Registrar"
#define DBUSMENU_REG_OBJECT "/com/canonical/AppMenu/Registrar"

G_BEGIN_DECLS

G_DECLARE_FINAL_TYPE(RegistrarDBusMenu, registrar_dbus_menu, REGISTRAR, DBUS_MENU, GObject)
uint registrar_dbus_menu_register(RegistrarDBusMenu *object, GDBusConnection *connection,
                                  GError **error);
void registrar_dbus_menu_unregister(RegistrarDBusMenu *data, GDBusConnection *con);

G_END_DECLS

#endif // REGISTRARDBUSMENU_H
